#include <Python.h>
#include <Numeric/arrayobject.h>
#include "mt.h"

#define MIN(a, b) ((a)<(b)?(a):(b))
#define MAX(a, b) ((a)>(b)?(a):(b))
#define ABS(a)    MAX((a),-(a))
#define GETVEC(m, x, y) ((m)+((x)*h+(y))*dim)


#define RAISEERROR(mesg) do { PyErr_SetString(PyExc_RuntimeError, (mesg)); return NULL; } while (0)

/*
 * mingw + ActivePython $B$G$bF0$+$;$k$h$&$K!"M>7W$J$b$N$O0l@Z(B include $B$7$J$$!#(B
 * I/O $B$^$o$j$O(B python $B7PM3$G9T$&!#(B
 */

/*
 * $B$*$*$`$M%(%i!<%A%'%C%/$O4K$$!#(B
 * $B0z?t$N(B array $B$O(B double $B$G(B contiguous $B$G$J$$$H!"2?$,5/$-$k$+$o$+$i$J$$!#(B
 */

PyObject *newNone(void) {
    PyObject *rv = Py_None;
    Py_INCREF(rv);
    return rv;
}

double mysqrt(double x) {
    double pv = 0;
    double tv = x;
    double deno = ABS(x);

    if (deno < 1e-9)
	return 0;

    while (ABS(pv-tv)/deno > 1e-5) {
	pv = tv;
	tv = (x / tv + tv) / 2;
    }
    return tv;
}

double norm2(double *a, double *b, int dim) {
    int i;
    double sum = 0;
    for (i=0; i<dim; i++)
	sum += (a[i]-b[i])*(a[i]-b[i]);
    return sum;
}

double norm(double *a, double *b, int dim) {
    return mysqrt(norm2(a, b, dim));
}

PyObject *randomize(PyObject *dmy, PyObject *args) {
    PyObject *o;
    PyArrayObject *a;
    double *dp, *lim;
    int i, n;

    if (!PyArg_ParseTuple(args, "O!", &PyArray_Type, &o)) {
	return NULL;
    }
    a = (PyArrayObject *)o;

    n = 1;
    for (i=0; i<a->nd; i++)
	n *= a->dimensions[i];
    dp = (double *)a->data;
    lim = dp + n;
    while (dp < lim)
	*dp++ = mtRandomDouble();
    
    return newNone();
}

PyObject *somcoreSrandom(PyObject *dmy, PyObject *args) {
    int i;
    if (!PyArg_ParseTuple(args, "i", &i))
	return NULL;
    mtSrandom(i);
    return newNone();
}

double distFactorBubble(double dx, double dy, double nr2) {
    return (dx*dx+dy*dy <= nr2) ? 1 : 0;
}

void updateBubble(double *a, double *v,
		  int w, int h, int dim,
		  int cx, int cy,
		  double decayFactor, double neighbourRadius) {
    int x, y, i;
    int mix, max, miy, may;
    double nr2;

    mix = MAX(0, cx-neighbourRadius-1);
    max = MIN(w-1, cx+neighbourRadius);
    miy = MAX(0, cy-neighbourRadius-1);
    may = MIN(h-1, cy+neighbourRadius);

    nr2 = neighbourRadius * neighbourRadius;

    for (x=mix; x<=max; x++) {
	for (y=miy; y<=may; y++) {
	    double e = decayFactor * distFactorBubble(ABS(cx-x), ABS(cy-y), nr2);
	    double *dp = a + (h*x+y)*dim;
	    for (i=0; i<dim; i++)
		dp[i] += e*(v[i]-dp[i]);
	}
    }
}

PyObject *update(PyObject *dmy, PyObject *args) {
    PyArrayObject *am, *vv;
    int cx, cy;
    double decayFactor, neighbourRadius;
    int w, h, dim;

    if (!PyArg_ParseTuple(args, "O!O!iidd",
			  &PyArray_Type, &am,
			  &PyArray_Type, &vv,
			  &cx, &cy, &decayFactor, &neighbourRadius))
	return NULL;

    if (am->nd != 3)
	RAISEERROR("Dimension not 3!");

    w = am->dimensions[0];
    h = am->dimensions[1];
    dim = am->dimensions[2];
    
    if (vv->nd != 1)
	RAISEERROR("Not a vector!");
    if (dim != vv->dimensions[0])
	RAISEERROR("dimension mismatch!");

    updateBubble((double *)am->data, (double *)vv->data, w, h, dim, cx, cy, decayFactor, neighbourRadius);
    return newNone();
}

int dX[] = {-1, -1, -1, 0, 1, 1, 1, 0};
int dY[] = { 1,  0, -1,-1,-1, 0, 1, 1};

void getWinnerCneighb(double *m, double *v, double *nm, int w, int h, int dim, int *xio, int *yio) {
    int x = *xio;
    int y = *yio;
    int i;
    double cv = norm2(v, GETVEC(m, x, y), dim);

    while (1) {
	int bx = -1;
	int by = -1;
	for (i=0; i<sizeof(dX)/sizeof(dX[0]); i++) {
	    int nx = x + dX[i];
	    int ny = y + dY[i];
	    if (nx >= 0 && nx < w && ny >= 0 && ny < h) {
		double nv = norm2(v, GETVEC(m, nx, ny), dim);
		if (nv < cv) {
		    bx = nx;
		    by = ny;
		    cv = nv;
		}
	    }
	}
	if (bx >= 0) {
	    x = bx;
	    y = by;
	} else {
	    *xio = x;
	    *yio = y;
	    return;
	}
    }
}

void getWinnerC(double *m, double *v, double *nm, int w, int h, int dim, int *xret, int *yret) {
    double *dp, *dvp, *dnmp, *dplim, *dvplim;
    double *wp;
    double minv;

    dp = m;
    dnmp = nm;
    dvplim = v + dim;
    dplim = m + dim * w * h;

    wp = m;
    minv = 1e+100;

    while (dp < dplim) {
	double tv = 0;
	dvp = v;
	while (dvp < dvplim) {
	    tv += (*dp-*dvp)*(*dp-*dvp);
	    dp++;
	    dvp++;
	}
	if (tv < minv) {
	    wp = dp-dim;
	    minv = tv;
	}
	if (dnmp != NULL) {
	    *dnmp++ = tv;
	}
    }

    *xret = (wp - m)/dim/h;
    *yret = (wp - m)/dim%h;
}
    
PyObject *getWinner(PyObject *dmy, PyObject *args) {
    PyArrayObject *am, *vv, *anm;
    double *m, *v, *nm;
    int w, h, dim;
    int wx, wy;
    int i;

    anm = NULL;
    if (!PyArg_ParseTuple(args, "O!O!|O!", &PyArray_Type, &am, &PyArray_Type, &vv, &PyArray_Type, &anm))
	return NULL;

    m = (double *)am->data;
    if (anm != NULL)
	nm = (double *)anm->data;
    else
	nm = NULL;
    v = (double *)vv->data;

    if (am->nd != 3)
	RAISEERROR("dimension not 3!");
    if (anm != NULL && anm->nd != 2)
	RAISEERROR("dimension not 3!");

    w = am->dimensions[0];
    h = am->dimensions[1];
    dim = am->dimensions[2];

    if (vv->dimensions[0] != dim)
	RAISEERROR("dimension mismatch!");

    if (anm != NULL)
	for (i=0; i<2; i++)
	    if (am->dimensions[i] != anm->dimensions[i])
		RAISEERROR("dimension mismatch! (m, nm)");

    getWinnerC(m, v, nm, w, h, dim, &wx, &wy);

    return Py_BuildValue("(ii)", wx, wy);
}

#define SETRGB(p, r, g, b) 

void doDisplay(double *m, unsigned char *s, double *nm,
	       int w, int h, double nmthresh, int mag, int dim) {
    double *mlim = m+w*h*dim;
    int x, y;

    x = y = 0;
    while (m < mlim) {
	unsigned char r, g, b, *cp2, *cp3, *cplim;
	int tv;
	int dx;

	r = m[6]*128;
	tv = (m[10]+m[11]+m[12])*128;
	g = MIN(tv, 255);
	tv = m[16]/1.5*256;
	b = MIN(tv, 255);

	if (*nm < nmthresh) {
	    double rat = *nm / nmthresh;
	    r = 255+(r-255.0)*rat;
	    g = 255+(g-255.0)*rat;
	    b = 255+(b-255.0)*rat;
	}
	m += dim;
	nm++;
	cp2 = s + (x * mag * h + y) * mag * 3;
	for (dx=0; dx<mag; dx++) {
	    cp3 = cp2 + dx*h*mag*3;
	    cplim = cp3 + mag * 3;
	    while (cp3 < cplim) {
		*cp3++ = r;
		*cp3++ = g;
		*cp3++ = b;
	    }
	}
	y++;
	if (y >= h) {
	    x++;
	    y = 0;
	}
    }
}

PyObject *displayInSurfArray(PyObject *dmy, PyObject *args) {
    PyArrayObject *am, *as, *anm;
    double nmthresh;
    int mag;
    int w, h, dim;

    if (!PyArg_ParseTuple(args, "O!O!O!di",
			  &PyArray_Type, &am,
			  &PyArray_Type, &as,
			  &PyArray_Type, &anm,
			  &nmthresh, &mag))
	return NULL;

    if (am->nd != 3 || as->nd != 3)
	RAISEERROR("dimension not 3!");
    if (anm->nd != 2)
	RAISEERROR("dimension not 3!");

    w = am->dimensions[0];
    h = am->dimensions[1];
    dim = am->dimensions[2];

    if (!(w*mag == as->dimensions[0] && h*mag == as->dimensions[1] &&
	  anm->dimensions[0] == w && anm->dimensions[1] == h))
	RAISEERROR("dimensional error!");

    if (as->dimensions[2] != 3)
	RAISEERROR("Sorry, displayInSurfArray supports 24bpp surfarray only");
/*     if (am->dimensions[2] != 3) */
/* 	RAISEERROR("Sorry, displayInSurfArray supports 3D SOM-source only"); */

    doDisplay((double *)am->data, (unsigned char *)as->data, (double *)anm->data, w, h, nmthresh, mag, dim);
    return newNone();
}

#define THINNINGFACTOR 10
static int gHitclass = 0;

void doCalcMappings(double *vecs, double *m, double *mappings,
		    int w, int h, int dim,
		    int N,
		    int accurateFlag) {
    int i;
    gHitclass = (gHitclass + 1)%THINNINGFACTOR;
    for (i=0; i<N; i++) {
	int x, y;
	double *v = vecs + i*dim;
	double n;
	double dx, dy;

	if (accurateFlag ||
	    i * THINNINGFACTOR / N == gHitclass) {
	    getWinnerC(m, v, NULL, w, h, dim, &x, &y);
	} else {
	    x = mappings[i*2];
	    y = mappings[i*2+1];
	    getWinnerCneighb(m, v, NULL, w, h, dim, &x, &y);
	}
	
	n = norm(v, GETVEC(m, x, y), dim);
	if (x == 0 || x == w-1) {
	    dx = 0;
	} else {
	    double nm = norm(v, GETVEC(m, x-1, y), dim)-n;
	    double np = norm(v, GETVEC(m, x+1, y), dim)-n;
	    if (nm + np == 0)
		dx = 0;
	    else
		dx = nm/(nm+np)-0.5;
	}

	if (y == 0 || y == h-1) {
	    dy = 0;
	} else {
	    double nm = norm(v, GETVEC(m, x, y-1), dim)-n;
	    double np = norm(v, GETVEC(m, x, y+1), dim)-n;
	    if (nm + np == 0)
		dy = 0;
	    else
		dy = nm/(nm+np)-0.5;
	}

	mappings[i*2] = x+dx;
	mappings[i*2+1] = y+dy;
    }
}


PyObject *calcMappings(PyObject *dmy, PyObject *args) {
    PyArrayObject *vvecs, *am, *amappings;
    int w, h, dim, n;
    double *vecs, *m, *mappings;
    int accurateFlag;

    if (!PyArg_ParseTuple(args, "O!O!O!i",
			  &PyArray_Type, &vvecs,
			  &PyArray_Type, &am,
			  &PyArray_Type, &amappings,
			  &accurateFlag))
	return NULL;
    w = am->dimensions[0];
    h = am->dimensions[1];
    dim =  am->dimensions[2];
    n = vvecs->dimensions[0];

    vecs = (double *)vvecs->data;
    m = (double *)am->data;
    mappings = (double *)amappings->data;

    doCalcMappings(vecs, m, mappings, w, h, dim, n, accurateFlag);

    return newNone();
}

static PyMethodDef methes[] = {
    {"srandom", (PyCFunction)somcoreSrandom, METH_VARARGS, NULL},
    {"randomize", (PyCFunction)randomize, METH_VARARGS, NULL},
    {"getWinner", (PyCFunction)getWinner, METH_VARARGS, NULL},
    {"update", (PyCFunction)update, METH_VARARGS, NULL},
    {"displayInSurfArray", (PyCFunction)displayInSurfArray, METH_VARARGS, NULL},
    {"calcMappings", (PyCFunction)calcMappings, METH_VARARGS, NULL},
    {0, 0, 0, 0}
};

void initsomcore(void) {
    PyObject *m;
    import_array();
    m = Py_InitModule("somcore", methes);
}
